@@ -19,9 +19,11 @@ from typing import Any, Dict, Iterable, List, Optional, Sequence, Union
 import numpy as np
 
 from fastestimator.dataset.dataset import DatasetSummary, FEDataset
+from fastestimator.util.traceability_util import traceable
 from fastestimator.util.util import to_list
 
 
+@traceable()
 class BatchDataset(FEDataset):
     """BatchDataset extracts a list (batch) of data from a single dataset or multiple datasets.
 
@@ -49,7 +51,8 @@ class BatchDataset(FEDataset):
         ```
 
     Args:
-        datasets: The dataset(s) to use for batch sampling.
+        datasets: The dataset(s) to use for batch sampling. While these should be FEDatasets, pytorch datasets will
+            technically also work. If you use them, however, you will lose the .split() and .summary() methods.
         num_samples: Number of samples to draw from the `datasets`. May be a single int if used in conjunction with
             `probability`, otherwise a list of ints of len(`datasets`) is required.
         probability: Probability to draw from each dataset. Only allowed if `num_samples` is an integer.
@@ -62,7 +65,9 @@ class BatchDataset(FEDataset):
         self.num_samples = to_list(num_samples)
         self.probability = to_list(probability)
         self.same_feature = False
+        self.all_fe_datasets = False
         self._check_input()
+        self.index_maps = []
         self.reset_index_maps()
         self.pad_value = None
 
@@ -96,8 +101,9 @@ class BatchDataset(FEDataset):
             assert len(self.datasets) == len(self.num_samples), "the number of dataset must match num_samples"
         if not self.same_feature:
             assert len(set(self.num_samples)) == 1, "the number of samples must be the same for disjoint features"
+        self.all_fe_datasets = all([isinstance(dataset, FEDataset) for dataset in self.datasets])
 
-    def _do_split(self, splits: Sequence[Iterable[int]]) -> List['UnpairedDataset']:
+    def _do_split(self, splits: Sequence[Iterable[int]]) -> List['BatchDataset']:
         """This class overwrites the .split() method instead of _do_split().
 
         Args:
@@ -109,7 +115,7 @@ class BatchDataset(FEDataset):
         """
         raise AssertionError("This method should not have been invoked. Please file a bug report")
 
-    def split(self, *fractions: Union[float, int, Iterable[int]]) -> Union['UnpairedDataset', List['UnpairedDataset']]:
+    def split(self, *fractions: Union[float, int, Iterable[int]]) -> Union['BatchDataset', List['BatchDataset']]:
         """Split this dataset into multiple smaller datasets.
 
         This function enables several types of splitting:
@@ -141,13 +147,20 @@ class BatchDataset(FEDataset):
             One or more new datasets which are created by removing elements from the current dataset. The number of
             datasets returned will be equal to the number of `fractions` provided. If only a single value is provided
             then the return will be a single dataset rather than a list of datasets.
+
+        Raises:
+            NotImplementedError: If the user created this dataset using one or more non-FEDataset inputs.
         """
+        if not self.all_fe_datasets:
+            raise NotImplementedError(
+                "BatchDataset.split() is not supported when BatchDataset contains non-FEDataset objects")
         new_datasets = [to_list(ds.split(*fractions)) for ds in self.datasets]
         num_splits = len(new_datasets[0])
         new_datasets = [[ds[i] for ds in new_datasets] for i in range(num_splits)]
         results = [BatchDataset(ds, self.num_samples, self.probability) for ds in new_datasets]
         # Re-compute personal variables
         self.reset_index_maps()
+        FEDataset.fix_split_traceabilty(self, results, fractions)
         # Unpack response if only a single split
         if len(results) == 1:
             results = results[0]
@@ -158,6 +171,9 @@ class BatchDataset(FEDataset):
         Returns:
             A summary representation of this dataset.
         """
+        if not self.all_fe_datasets:
+            print("FastEstimator-Warn: BatchDataset summary will be incomplete since non-FEDatasets were used.")
+            return DatasetSummary(num_instances=len(self), keys={})
         summaries = [ds.summary() for ds in self.datasets]
         keys = {k: v for summary in summaries for k, v in summary.keys.items()}
         return DatasetSummary(num_instances=len(self), keys=keys)
