@@ -15,19 +15,23 @@
 import math
 import random
 from collections import defaultdict
+from copy import deepcopy
 from functools import lru_cache
-from typing import Any, Dict, Hashable, Iterable, List, Optional, Sequence, Union
+from typing import Any, Dict, Hashable, Iterable, List, Optional, Sequence, Tuple, Union
 
 import jsonpickle
 import numpy as np
 from torch.utils.data import Dataset
 
-from fastestimator.util.util import get_shape, get_type
+from fastestimator.util.traceability_util import FeSplitSummary, traceable
+from fastestimator.util.util import FEID, get_shape, get_type
 
 
 class KeySummary:
     """A summary of the dataset attributes corresponding to a particular key.
 
+    This class is intentionally not @traceable.
+
     Args:
         num_unique_values: The number of unique values corresponding to a particular key (if known).
         shape: The shape of the vectors corresponding to the key. None is used in a list to indicate that a dimension is
@@ -53,6 +57,8 @@ class KeySummary:
 class DatasetSummary:
     """This class contains information summarizing a dataset object.
 
+    This class is intentionally not @traceable.
+
     Args:
         num_instances: The number of data instances within the dataset (influences the size of an epoch).
         num_classes: How many different classes are present.
@@ -88,6 +94,7 @@ class DatasetSummary:
         return jsonpickle.dumps(self, unpicklable=False)
 
 
+@traceable()
 class FEDataset(Dataset):
     def __len__(self) -> int:
         """Defines how many datapoints the dataset contains.
@@ -110,6 +117,53 @@ class FEDataset(Dataset):
         """
         raise NotImplementedError
 
+    @classmethod
+    def fix_split_traceabilty(cls,
+                              parent: 'FEDataset',
+                              children: List['FEDataset'],
+                              fractions: Tuple[Union[float, int, Iterable[int]], ...]) -> None:
+        """A method to fix traceability information after invoking the dataset .split() method.
+
+        Note that the default implementation of the .split() function invokes this already, so this only needs to be
+        invoked if you override the .split() method when defining a subclass (ex. BatchDataset).
+
+        Args:
+            parent: The parent dataset on which .split() was invoked.
+            children: The datasets generated by performing the split.
+            fractions: The fraction arguments used to generate the children (should be one-to-one with the children).
+        """
+        if hasattr(parent, '_fe_traceability_summary'):
+            parent_id = FEID(id(parent))
+            fractions = [
+                f"range({frac.start}, {frac.stop}, {frac.step})" if isinstance(frac, range) else f"{frac}"
+                for frac in fractions
+            ]
+            for child, frac in zip(children, fractions):
+                # noinspection PyProtectedMember
+                tables = deepcopy(child._fe_traceability_summary)
+                # Update the ID if necessary
+                child_id = FEID(id(child))
+                if child_id not in tables:
+                    # The child was created without invoking its __init__ method, so its internal summary will have the
+                    # wrong id
+                    table = tables.pop(parent_id)
+                    table.fe_id = child_id
+                    tables[child_id] = table
+                else:
+                    table = tables[child_id]
+                split_summary = table.fields.get('split', FeSplitSummary())
+                split_summary.add_split(parent=parent_id, fraction=frac)
+                table.fields['split'] = split_summary
+                child._fe_traceability_summary = tables
+            # noinspection PyUnresolvedReferences
+            table = parent._fe_traceability_summary.get(parent_id)
+            split_summary = table.fields.get('split', FeSplitSummary())
+            split_summary.add_split(parent='self', fraction=", ".join([f"-{frac}" for frac in fractions]))
+            table.fields['split'] = split_summary
+            # Put the new parent summary into the child table to ensure it will always exist in the final set of tables
+            for child in children:
+                child._fe_traceability_summary[parent_id] = deepcopy(table)
+
     def split(self, *fractions: Union[float, int, Iterable[int]]) -> Union['FEDataset', List['FEDataset']]:
         """Split this dataset into multiple smaller datasets.
 
@@ -127,7 +181,7 @@ class FEDataset(Dataset):
             ds3, ds4 = ds.split(90, 180)  # len(ds) == 630, len(ds3) == 90, len(ds4) == 180
             ```
         3. Splitting by indices.
-            ``python
+            ```python
             ds = fe.dataset.FEDataset(...)  # len(ds) == 1000
             ds2 = ds.split([87,2,3,100,121,158])  # len(ds) == 994, len(ds2) == 6
             ds3 = ds.split(range(100))  # len(ds) == 894, len(ds3) == 100
@@ -187,6 +241,7 @@ class FEDataset(Dataset):
         elif method == 'indices':
             splits = fractions
         splits = self._do_split(splits)
+        FEDataset.fix_split_traceabilty(self, splits, fractions)
         if len(fractions) == 1:
             return splits[0]
         return splits
@@ -224,6 +279,7 @@ class FEDataset(Dataset):
         return str(self.summary())
 
 
+@traceable()
 class InMemoryDataset(FEDataset):
     """A dataset abstraction to simplify the implementation of datasets which hold their data in memory.
 
