@@ -14,7 +14,7 @@
 # ==============================================================================
 import os
 import random
-from collections import ChainMap, deque
+from collections import ChainMap
 from typing import Any, Dict, Iterable, List, Optional, Set, Union
 
 import numpy as np
@@ -23,6 +23,7 @@ import torch
 from tensorflow.python.distribute.input_lib import DistributedDataset
 from torch.utils.data import DataLoader
 
+import fastestimator as fe
 from fastestimator.backend.to_shape import to_shape
 from fastestimator.backend.to_tensor import to_tensor
 from fastestimator.backend.to_type import to_type
@@ -33,11 +34,14 @@ from fastestimator.schedule.schedule import Scheduler, get_current_items, get_si
 from fastestimator.summary.system import Summary, System
 from fastestimator.trace.io.best_model_saver import BestModelSaver
 from fastestimator.trace.io.model_saver import ModelSaver
-from fastestimator.trace.trace import EvalEssential, Logger, Trace, TrainEssential
+from fastestimator.trace.io.traceability import Traceability
+from fastestimator.trace.trace import EvalEssential, Logger, Trace, TrainEssential, sort_traces
 from fastestimator.util.data import Data
+from fastestimator.util.traceability_util import traceable
 from fastestimator.util.util import Suppressor, draw, to_list, to_set
 
 
+@traceable()
 class Estimator:
     """One class to rule them all.
 
@@ -58,9 +62,9 @@ class Estimator:
             information will still print). None to completely disable printing.
         monitor_names: Additional keys from the data dictionary to be written into the logs.
     """
-    pipeline: Pipeline
-    traces: List[Union[Trace, Scheduler[Trace]]]
     monitor_names: Set[str]
+    traces_in_use: List[Union[Trace, Scheduler[Trace]]]
+    system: System
 
     def __init__(self,
                  pipeline: Pipeline,
@@ -71,20 +75,32 @@ class Estimator:
                  traces: Union[None, Trace, Scheduler[Trace], Iterable[Union[Trace, Scheduler[Trace]]]] = None,
                  log_steps: Optional[int] = 100,
                  monitor_names: Union[None, str, Iterable[str]] = None):
-        self.pipeline = pipeline
-        self.network = network
-        self.traces = to_list(traces)
-        self.traces_in_use = None
+        self.traces_in_use = []
         assert log_steps is None or log_steps >= 0, \
             "log_steps must be None or positive (or 0 to disable only train logging)"
-        self.monitor_names = to_set(monitor_names) | self.network.get_loss_keys()
+        self.monitor_names = to_set(monitor_names) | network.get_loss_keys()
         self.system = System(network=network,
+                             pipeline=pipeline,
+                             traces=to_list(traces),
                              log_steps=log_steps,
                              total_epochs=epochs,
                              max_train_steps_per_epoch=max_train_steps_per_epoch,
-                             max_eval_steps_per_epoch=max_eval_steps_per_epoch)
+                             max_eval_steps_per_epoch=max_eval_steps_per_epoch,
+                             system_config=self.fe_summary())
 
-    def fit(self, summary: Optional[str] = None, warmup: bool = True) -> Optional[Summary]:
+    @property
+    def pipeline(self) -> Pipeline:
+        return self.system.pipeline
+
+    @property
+    def network(self) -> BaseNetwork:
+        return self.system.network
+
+    @property
+    def traces(self) -> List[Union[Trace, Scheduler[Trace]]]:
+        return self.system.traces
+
+    def fit(self, summary: Optional[str] = None, warmup: Union[bool, str] = True) -> Optional[Summary]:
         """Train the network for the number of epochs specified by the estimator's constructor.
 
         Args:
@@ -93,16 +109,16 @@ class Estimator:
             warmup: Whether to perform warmup before training begins. The warmup procedure will test one step at every
                 epoch where schedulers cause the execution graph to change. This can take some time up front, but can
                 also save significant heartache on epoch 300 when the training unexpectedly fails due to a tensor size
-                mismatch.
+                mismatch. When set to "debug", the warmup will be performed in eager execution for easier debugging.
 
         Returns:
             A summary object containing the training history for this session iff a `summary` name was provided.
         """
         draw()
-        self.system.reset(summary)
+        self.system.reset(summary, self.fe_summary())
         self._prepare_traces(run_modes={"train", "eval"})
         if warmup:
-            self._warmup()
+            self._warmup(warmup=warmup)
         self._start(run_modes={"train", "eval"})
         return self.system.summary or None
 
@@ -117,8 +133,15 @@ class Estimator:
         self.traces_in_use = [trace for trace in self.traces]
         if self.system.log_steps is not None:
             self.traces_in_use.append(Logger())
+        # Look for any monitor names which should be automagically added.
+        trace_outputs = set()
+        extra_monitor_keys = set()
+        for trace in sort_traces(get_current_items(self.traces_in_use, run_modes=run_modes)):
+            trace_outputs.update(trace.outputs)
+            extra_monitor_keys.update(trace.fe_monitor_names - trace_outputs)
+        # Add the essential traces
         if "train" in run_modes:
-            self.traces_in_use.insert(0, TrainEssential(monitor_names=self.monitor_names))
+            self.traces_in_use.insert(0, TrainEssential(monitor_names=self.monitor_names.union(extra_monitor_keys)))
             no_save_warning = True
             for trace in get_current_items(self.traces_in_use, run_modes=run_modes):
                 if isinstance(trace, (ModelSaver, BestModelSaver)):
@@ -126,7 +149,7 @@ class Estimator:
             if no_save_warning:
                 print("FastEstimator-Warn: No ModelSaver Trace detected. Models will not be saved.")
         if "eval" in run_modes and "eval" in self.pipeline.get_modes():
-            self.traces_in_use.insert(1, EvalEssential(monitor_names=self.monitor_names))
+            self.traces_in_use.insert(1, EvalEssential(monitor_names=self.monitor_names.union(extra_monitor_keys)))
         # insert system instance to trace
         for trace in get_current_items(self.traces_in_use, run_modes=run_modes):
             trace.system = self.system
@@ -148,87 +171,17 @@ class Estimator:
         self._start(run_modes={"test"})
         return self.system.summary or None
 
-    @staticmethod
-    def _sort_traces(traces: List[Trace], available_outputs: Optional[Set[str]] = None) -> List[Trace]:
-        """Sort traces to attempt to resolve any dependency issues.
-
-        This is essentially a topological sort, but it doesn't seem worthwhile to convert the data into a graph
-        representation in order to get the slightly better asymptotic runtime complexity.
-
-        Args:
-            traces: A list of traces (not inside schedulers) to be sorted.
-            available_outputs: What output keys are already available for the traces to use. If None are provided, the
-                sorting algorithm will assume that any keys not generated by traces are being provided by the system.
-                This results in a less rigorous sorting.
-
-        Returns:
-            The sorted list of `traces`.
-
-        Raises:
-            AssertionError: If Traces have circular dependencies or require input keys which are not available.
-        """
-        sorted_traces = []
-        trace_outputs = {output for trace in traces for output in trace.outputs}
-        if available_outputs is None:
-            # Assume that anything not generated by a Trace is provided by the system
-            available_outputs = {inp for trace in traces for inp in trace.inputs} - trace_outputs
-            weak_sort = True
-        else:
-            available_outputs = to_set(available_outputs)
-            weak_sort = False
-        end_traces = deque()
-        intermediate_traces = deque()
-        intermediate_outputs = set()
-        trace_deque = deque(traces)
-        while trace_deque:
-            trace = trace_deque.popleft()
-            ins = set(trace.inputs)
-            outs = set(trace.outputs)
-            if not ins or isinstance(trace, (TrainEssential, EvalEssential)):
-                sorted_traces.append(trace)
-                available_outputs |= outs
-            elif "*" in ins:
-                if outs:
-                    end_traces.appendleft(trace)
-                else:
-                    end_traces.append(trace)
-            elif ins <= available_outputs or (weak_sort and (ins - outs - available_outputs).isdisjoint(trace_outputs)):
-                sorted_traces.append(trace)
-                available_outputs |= outs
-            else:
-                intermediate_traces.append(trace)
-                intermediate_outputs |= outs
-
-        already_seen = set()
-        while intermediate_traces:
-            trace = intermediate_traces.popleft()
-            ins = set(trace.inputs)
-            outs = set(trace.outputs)
-            already_seen.add(trace)
-            if ins <= available_outputs or (weak_sort and (ins - outs - available_outputs).isdisjoint(trace_outputs)):
-                sorted_traces.append(trace)
-                available_outputs |= outs
-                already_seen.clear()
-            elif ins <= (available_outputs | intermediate_outputs):
-                intermediate_traces.append(trace)
-            else:
-                raise AssertionError("The {} trace has unsatisfiable inputs: {}".format(
-                    type(trace).__name__, ", ".join(ins - (available_outputs | intermediate_outputs))))
-
-            if intermediate_traces and len(already_seen) == len(intermediate_traces):
-                raise AssertionError("Dependency cycle detected amongst traces: {}".format(", ".join(
-                    [type(tr).__name__ for tr in already_seen])))
-        sorted_traces.extend(list(end_traces))
-        return sorted_traces
-
-    def _warmup(self) -> None:
+    def _warmup(self, warmup: Union[bool, str]) -> None:
         """Perform a test run of each pipeline and network signature epoch to make sure that training won't fail later.
 
         Traces are not executed in the warmup since they are likely to contain state variables which could become
         corrupted by running extra steps.
+
+        Args:
+            warmup: Warmup arg specified by estimator.fit.
         """
         all_traces = get_current_items(self.traces_in_use, run_modes={"train", "eval"})
-        self._sort_traces(all_traces)
+        sort_traces(all_traces)  # This ensures that the traces can sort properly for on_begin and on_end
         monitor_names = self.monitor_names
         for mode in self.pipeline.get_modes() - {"test"}:
             scheduled_items = self.pipeline.get_scheduled_items(mode) + self.network.get_scheduled_items(
@@ -260,9 +213,9 @@ class Estimator:
                 unmet_requirements = trace_input_keys - (pipeline_output_keys | network_output_keys | trace_output_keys)
                 assert not unmet_requirements, \
                     "found missing key(s) during epoch {} mode {}: {}".format(epoch, mode, unmet_requirements)
-                self._sort_traces(traces, available_outputs=pipeline_output_keys | network_output_keys)
+                sort_traces(traces, available_outputs=pipeline_output_keys | network_output_keys)
                 trace_input_keys.update(traces[0].inputs)
-                self.network.load_epoch(mode, epoch, output_keys=trace_input_keys, warmup=True)
+                self.network.load_epoch(mode, epoch, output_keys=trace_input_keys, warmup=warmup)
                 self.network.run_step(batch)
                 self.network.unload_epoch()
         assert not monitor_names, "found missing key(s): {}".format(monitor_names)
@@ -287,10 +240,9 @@ class Estimator:
         Args:
             run_modes: The current execution modes.
         """
-        all_traces = get_current_items(self.traces_in_use, run_modes=run_modes)
-        self._sort_traces(all_traces)
-        self._run_traces_on_begin(traces=all_traces)
+        all_traces = sort_traces(get_current_items(self.traces_in_use, run_modes=run_modes))
         try:
+            self._run_traces_on_begin(traces=all_traces)
             if "train" in run_modes or "eval" in run_modes:
                 for self.system.epoch_idx in range(self.system.epoch_idx + 1, self.system.total_epochs + 1):
                     if "train" in self.pipeline.get_modes(epoch=self.system.epoch_idx):
@@ -320,7 +272,7 @@ class Estimator:
         self.system.batch_idx = None
         with Suppressor():
             batch = next(iterator)
-        traces = self._sort_traces(
+        traces = sort_traces(
             traces,
             available_outputs=to_set(batch.keys())
             | self.network.get_all_output_keys(self.system.mode, self.system.epoch_idx))
@@ -330,8 +282,8 @@ class Estimator:
                 if self.system.mode == "train":
                     self.system.update_global_step()
                 self.system.update_batch_idx()
-                self._run_traces_on_batch_begin(traces=traces)
                 batch = self._configure_tensor(loader, batch)
+                self._run_traces_on_batch_begin(batch, traces=traces)
                 batch, prediction = self.network.run_step(batch)
                 self._run_traces_on_batch_end(batch, prediction, traces=traces)
                 if isinstance(loader, DataLoader) and (
@@ -409,13 +361,14 @@ class Estimator:
             trace.on_epoch_begin(data)
         self._check_early_exit()
 
-    def _run_traces_on_batch_begin(self, traces: Iterable[Trace]) -> None:
+    def _run_traces_on_batch_begin(self, batch: Dict[str, Any], traces: Iterable[Trace]) -> None:
         """Invoke the on_batch_begin methods of given traces.
 
         Args:
+            batch: The batch data which was provided by the pipeline.
             traces: List of traces.
         """
-        data = Data()
+        data = Data(batch)
         for trace in traces:
             trace.on_batch_begin(data)
         self._check_early_exit()
@@ -453,8 +406,15 @@ class Estimator:
             traces: List of traces.
         """
         data = Data()
+        traceability = None
         for trace in traces:
+            if isinstance(trace, Traceability):
+                # Delay traceability until the end so that it can capture all data including the total training time
+                traceability = trace
+                continue
             trace.on_end(data)
+        if traceability:
+            traceability.on_end(data)
 
     def _check_early_exit(self) -> None:
         """Determine whether training should be prematurely aborted.
@@ -468,10 +428,20 @@ class Estimator:
 
 class EarlyStop(Exception):
     """An exception raised when the system.stop_training flag is flipped by a Trace in order to abort the training.
+
+    This class is intentionally not @traceable.
     """
 
 
 def enable_deterministic(seed):
+    """Invoke to set random seed for deterministic training. The determinism only works for tensorflow >= 2.1 and
+    pytorch >= 1.14, and some model layers don't support.
+
+    Known failing layers:
+    * tf.keras.layers.UpSampling2D
+
+    """
+    fe.fe_deterministic_seed = seed
     os.environ['PYTHONHASHSEED'] = str(seed)
     os.environ['TF_DETERMINISTIC_OPS'] = str(1)
     random.seed(seed)
