@@ -12,26 +12,33 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 # ==============================================================================
+import os
+import tempfile
 from collections import ChainMap
 from typing import Any, Callable, Dict, Iterable, List, MutableMapping, Optional, Set, Tuple, TypeVar, Union
 
+import gdown
 import tensorflow as tf
 import torch
+from tensorflow.keras.mixed_precision import experimental as mixed_precision_tf
 from tensorflow.python.distribute.values import DistributedValues
 
 from fastestimator.backend.load_model import load_model
 from fastestimator.backend.to_tensor import to_tensor
 from fastestimator.op.op import get_inputs_by_op, write_outputs_by_op
 from fastestimator.op.tensorop import TensorOp
-from fastestimator.op.tensorop.model.model import ModelOp
-from fastestimator.op.tensorop.model.update import UpdateOp
+from fastestimator.op.tensorop.model import UpdateOp
 from fastestimator.schedule.schedule import EpochScheduler, RepeatScheduler, Scheduler, get_current_items
+from fastestimator.util.traceability_util import trace_model, traceable
 from fastestimator.util.util import NonContext, get_batch_size, to_list
 
 Model = TypeVar('Model', tf.keras.Model, torch.nn.Module)
 T = TypeVar('T')
 
+GOOGLE_DRIVE_URL = "https://drive.google.com"
 
+
+@traceable()
 class BaseNetwork:
     """A base class for Network objects.
 
@@ -50,6 +57,7 @@ class BaseNetwork:
         self.epoch_ops = []
         self.epoch_models = set()
         self.epoch_state = dict()
+        self.scaler = None
 
     def _verify_inputs(self) -> None:
         """Ensure that all ops are TensorOps.
@@ -92,11 +100,19 @@ class BaseNetwork:
         if output_keys:
             self.effective_outputs[mode] = self.effective_outputs[mode].intersection(output_keys)
         self.epoch_ops = get_current_items(self.ops, mode, epoch)
-        self.epoch_models = set(op.model for op in self.epoch_ops if isinstance(op, (UpdateOp, ModelOp)))
-        gradient_ops = [op for op in self.epoch_ops if hasattr(op, "retain_graph")]
+        self.epoch_models = set.union(*[op.get_fe_models() for op in self.epoch_ops])
+        gradient_ops = [op for op in self.epoch_ops if op.fe_retain_graph() is not None]
         for idx, gradient_op in enumerate(gradient_ops):
-            gradient_op.retain_graph = idx != len(gradient_ops) - 1
-        self.epoch_state = {"warmup": warmup, "mode": mode, "req_grad": len(gradient_ops) > 0}
+            gradient_op.fe_retain_graph(idx != len(gradient_ops) - 1)
+        self.epoch_state = {
+            "warmup": warmup,
+            "mode": mode,
+            "req_grad": len(gradient_ops) > 0,
+            "epoch": epoch,
+            "deferred": {},
+            "scaler": self.scaler
+        }
+        # warmup: bool, mode: str, req_grad: bool, epoch: int, deferred: Dict[str, List[Callable]]]
         for model in self.epoch_models:
             if hasattr(model, "optimizer") and model.optimizer is not None:
                 if isinstance(model.optimizer, Scheduler):
@@ -117,8 +133,7 @@ class BaseNetwork:
         """
         loss_keys = set()
         for op in get_current_items(self.ops):
-            if isinstance(op, UpdateOp):
-                loss_keys.update(op.inputs)
+            loss_keys |= op.get_fe_loss_keys()
         return loss_keys
 
     def get_effective_input_keys(self, mode: str, epoch: int) -> Set[str]:
@@ -168,6 +183,10 @@ class BaseNetwork:
             data = op.forward(data, state)
             if op.outputs:
                 write_outputs_by_op(op, batch, data)
+        for fn_list in state['deferred'].values():
+            for fn in fn_list:
+                fn()
+        state['deferred'].clear()
 
     def run_step(self, batch: Dict[str, Any]) -> Tuple[Dict[str, Any], Dict[str, Any]]:  # Batch, Prediction
         """Run a forward step through the Network on a batch of data.
@@ -208,8 +227,7 @@ def _collect_models(ops: Iterable[Union[TensorOp, Scheduler[TensorOp]]]) -> Set[
     """
     models = set()
     for op in get_current_items(ops):
-        if isinstance(op, (ModelOp, UpdateOp)):
-            models.add(op.model)
+        models |= op.get_fe_models()
     return models
 
 
@@ -230,16 +248,20 @@ def Network(ops: Iterable[Union[TensorOp, Scheduler[TensorOp]]]) -> BaseNetwork:
         ValueError: If a model is provided whose type cannot be identified as either TensorFlow or PyTorch.
     """
     models = _collect_models(ops)
-    assert models, "cannot find model in Network ops"
     framework = set()
+    model_names = set()
     for model in models:
+        model_names.add(model.model_name)
         if isinstance(model, tf.keras.Model):
             framework.add("tf")
         elif isinstance(model, torch.nn.Module):
             framework.add("torch")
         else:
             framework.add("unknown")
+    if len(framework) == 0:
+        framework.add('tf')  # We will use tf as default framework if no models are found
     assert len(framework) == 1, "please make sure either tensorflow or torch model is used in network"
+    assert len(model_names) == len(models), "all models must have unique model names"
 
     framework = framework.pop()
     if framework == "tf":
@@ -251,6 +273,7 @@ def Network(ops: Iterable[Union[TensorOp, Scheduler[TensorOp]]]) -> BaseNetwork:
     return network
 
 
+@traceable()
 class TorchNetwork(BaseNetwork):
     """An extension of BaseNetwork for PyTorch models.
 
@@ -259,7 +282,11 @@ class TorchNetwork(BaseNetwork):
     """
     def __init__(self, ops: Iterable[Union[TensorOp, Scheduler[TensorOp]]]) -> None:
         super().__init__(ops)
+        for op in get_current_items(self.ops):
+            op.build(framework='torch')
         self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
+        if any([model.mixed_precision for model in self.models]):
+            self.scaler = torch.cuda.amp.GradScaler()
 
     def load_epoch(self, mode: str, epoch: int, output_keys: Optional[Set[str]] = None, warmup: bool = False) -> None:
         """Prepare the network to run a given epoch and mode.
@@ -278,10 +305,19 @@ class TorchNetwork(BaseNetwork):
             for model in self.epoch_models:
                 # move model variables to gpu
                 model.to(self.device)
-                # move optimizer variables to gpu
-                self._move_optimizer_between_device(model.current_optimizer.state, self.device)
+                if model.current_optimizer and mode == "train":
+                    # move optimizer variables to gpu
+                    self._move_optimizer_between_device(model.current_optimizer.state, self.device)
+        # Set all of the contiguous final updates to defer their updates by default to enable things like CycleGan
+        # This is not necessary for TF because overriding tf weights does not confuse the gradient tape computation
+        for op in reversed(self.epoch_ops):
+            if isinstance(op, UpdateOp):
+                op._old_defer = op.defer
+                op.defer = True
+            else:
+                break
 
-    def _move_optimizer_between_device(self, data: Dict[str, Any], device: str) -> None:
+    def _move_optimizer_between_device(self, data: Dict[str, Any], device: Union[str, torch.device]) -> None:
         """Move optimizer state between gpu and cpu recursively.
 
         Args:
@@ -306,8 +342,15 @@ class TorchNetwork(BaseNetwork):
             for model in self.epoch_models:
                 # move model variables to cpu
                 model.to("cpu")
-                # move optimizer variables to cpu
-                self._move_optimizer_between_device(model.current_optimizer.state, "cpu")
+                if model.current_optimizer and self.epoch_state["mode"] == "train":
+                    # move optimizer variables to cpu
+                    self._move_optimizer_between_device(model.current_optimizer.state, "cpu")
+        # Set the final update ops back to their original defer status
+        for op in reversed(self.epoch_ops):
+            if isinstance(op, UpdateOp):
+                op.defer = op.__dict__.get('_old_defer', op.defer)
+            else:
+                break
 
     def _get_effective_batch_input(self, batch: MutableMapping[str, Any], mode: str) -> Dict[str, Any]:
         """Copy input data from the the CPU onto the GPU(s).
@@ -348,7 +391,12 @@ class TorchNetwork(BaseNetwork):
         self.epoch_state["tape"] = NonContext()
         # gpu operation
         with torch.no_grad() if not self.epoch_state["req_grad"] else NonContext():
-            self._forward_batch(batch_in, self.epoch_state, self.epoch_ops)
+            with torch.cuda.amp.autocast() if self.epoch_state["scaler"] is not None else NonContext():
+                self._forward_batch(batch_in, self.epoch_state, self.epoch_ops)
+        # if the loss scaler is used for training, update the scaler
+        if not self.epoch_state["warmup"] and self.epoch_state[
+                "mode"] == "train" and self.epoch_state["scaler"] is not None:
+            self.epoch_state["scaler"].update()
         # copy data to cpu
         if self.device.type == "cuda":
             prediction = {
@@ -362,7 +410,7 @@ class TorchNetwork(BaseNetwork):
             }
         return batch, prediction
 
-    def _move_tensor_between_device(self, data: T, device: str) -> T:
+    def _move_tensor_between_device(self, data: T, device: Union[str, torch.device]) -> T:
         """Move tensor between gpu and cpu recursively.
 
         Args:
@@ -416,7 +464,7 @@ class TorchNetwork(BaseNetwork):
         Returns:
             (batch_data, prediction_data)
         """
-        self.load_epoch(mode, epoch, warmup=True)
+        self.load_epoch(mode, epoch, warmup=False)
         data = to_tensor(data, "torch")
         data, prediction = self.run_step(data)
         self.unload_epoch()
@@ -424,9 +472,33 @@ class TorchNetwork(BaseNetwork):
         return data
 
 
+@traceable()
 class TFNetwork(BaseNetwork):
     """An extension of BaseNetwork for TensorFlow models.
+
+    Args:
+        ops: The ops defining the execution graph for this Network.
     """
+    def __init__(self, ops: Iterable[Union[TensorOp, Scheduler[TensorOp]]]) -> None:
+        super().__init__(ops)
+        for op in get_current_items(self.ops):
+            op.build(framework='tf')
+
+    def load_epoch(self, mode: str, epoch: int, output_keys: Optional[Set[str]] = None, warmup: bool = False) -> None:
+        """Prepare the network to run a given epoch and mode.
+
+        This method is necessary since schedulers and op mode restrictions may result in different computation graphs
+        every epoch. This also converts the epoch index a tensor to avoid tensorflow graph rebuilding.
+
+        Args:
+            mode: The mode to prepare to execute. One of 'train', 'eval', 'test', or 'infer'.
+            epoch: The epoch to prepare to execute.
+            output_keys: What keys must be moved from the GPU back to the CPU after executing a step.
+            warmup: Whether to prepare to execute it warmup mode or not (end users can likely ignore this argument).
+        """
+        super().load_epoch(mode, epoch, output_keys, warmup)
+        self.epoch_state["epoch"] = tf.convert_to_tensor(self.epoch_state["epoch"])
+
     def run_step(self, batch: Dict[str, Any]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
         """Run a forward step through the Network on a batch of data.
 
@@ -443,18 +515,18 @@ class TFNetwork(BaseNetwork):
         batch_in = self._get_effective_batch_input(batch, mode)
         strategy = tf.distribute.get_strategy()
         if isinstance(strategy, tf.distribute.MirroredStrategy):
-            if self.epoch_state["warmup"]:
-                prediction = strategy.experimental_run_v2(
+            if self.epoch_state["warmup"] == "debug":
+                prediction = strategy.run(
                     self._forward_step_eager,
                     args=(batch_in, self.epoch_state, self.epoch_ops, to_list(self.effective_outputs[mode])))
             else:
-                prediction = strategy.experimental_run_v2(
+                prediction = strategy.run(
                     self._forward_step_static,
                     args=(batch_in, self.epoch_state, self.epoch_ops, to_list(self.effective_outputs[mode])))
             batch = self._per_replica_to_global(batch)
             prediction = self._per_replica_to_global(prediction)
         else:
-            if self.epoch_state["warmup"]:
+            if self.epoch_state["warmup"] == "debug":
                 prediction = self._forward_step_eager(batch_in,
                                                       self.epoch_state,
                                                       self.epoch_ops,
@@ -560,7 +632,7 @@ class TFNetwork(BaseNetwork):
         Returns:
             The prediction dictionary resulting from a forward pass of the Network.
         """
-        batch = ChainMap({}, batch)
+        batch = dict(batch)
         prediction = {}
         with tf.GradientTape(persistent=True) if state["req_grad"] else NonContext() as tape:
             state['tape'] = tape
@@ -583,7 +655,7 @@ class TFNetwork(BaseNetwork):
         Returns:
             (batch_data, prediction_data)
         """
-        self.load_epoch(mode, epoch, warmup=True)
+        self.load_epoch(mode, epoch, warmup=False)
         data = to_tensor(data, target_type="tf")
         data, prediction = self.run_step(data)
         self.unload_epoch()
@@ -619,7 +691,8 @@ class TFNetwork(BaseNetwork):
 def build(model_fn: Callable[[], Union[Model, List[Model]]],
           optimizer_fn: Union[str, Scheduler, Callable, List[str], List[Callable], List[Scheduler], None],
           weights_path: Union[str, None, List[Union[str, None]]] = None,
-          model_name: Union[str, List[str], None] = None) -> Union[Model, List[Model]]:
+          model_name: Union[str, List[str], None] = None,
+          mixed_precision: bool = False) -> Union[Model, List[Model]]:
     """Build model instances and associate them with optimizers.
 
     This method can be used with TensorFlow models / optimizers:
@@ -646,6 +719,7 @@ def build(model_fn: Callable[[], Union[Model, List[Model]]],
             automatically generated and assigned.
         weights_path: Path(s) from which to load model weights. If not None, then the number of weight paths provided
             should match the number of models generated by the `model_fn`.
+        mixed_precision: Whether to enable mix precision network operations.
 
     Returns:
         models: The model(s) built by FastEstimator.
@@ -664,6 +738,11 @@ def build(model_fn: Callable[[], Union[Model, List[Model]]],
     # check framework
     if isinstance(models[0], tf.keras.Model):
         framework = "tf"
+        # tensorflow mix-precision instantiation
+        if mixed_precision:
+            mixed_precision_tf.set_policy(mixed_precision_tf.Policy('mixed_float16'))
+        else:
+            mixed_precision_tf.set_policy(mixed_precision_tf.Policy('float32'))
     elif isinstance(models[0], torch.nn.Module):
         framework = "torch"
     else:
@@ -675,6 +754,9 @@ def build(model_fn: Callable[[], Union[Model, List[Model]]],
             models = to_list(model_fn())
         if framework == "torch":
             models = [torch.nn.DataParallel(model) for model in models]
+    # mark models with its mixed_precision flag
+    for model in models:
+        model.mixed_precision = mixed_precision
     # generate names
     if not model_name:
         model_name = _generate_model_names(len(models))
@@ -688,7 +770,11 @@ def build(model_fn: Callable[[], Union[Model, List[Model]]],
         "Found inconsistency in number of models, optimizers, model_name or weights"
     # create optimizer
     for idx, (model, optimizer_def, weight, name) in enumerate(zip(models, optimizer_fn, weights_path, model_name)):
-        models[idx] = _fe_compile(model, optimizer_def, weight, name, framework)
+        models[idx] = trace_model(_fe_compile(model, optimizer_def, weight, name, framework),
+                                  model_idx=idx if len(models) > 1 else -1,
+                                  model_fn=model_fn,
+                                  optimizer_fn=optimizer_def,
+                                  weights_path=weight)
     if len(models) == 1:
         models = models[0]
     return models
@@ -725,8 +811,16 @@ def _fe_compile(model: Model,
         model.current_optimizer = optimizer_fn
     model.optimizer = optimizer_fn
     model.fe_compiled = True
+
     if weight:
+        if weight.startswith(GOOGLE_DRIVE_URL):
+            tmp_dir = tempfile.mkdtemp()
+            file_name = gdown.download(weight, quiet=False)
+            os.rename(os.path.join('./', file_name), os.path.join(tmp_dir, file_name))
+            weight = gdown.download(weight, os.path.join(tmp_dir, file_name), quiet=False)
+
         load_model(model, weight)
+
     model.model_name = name
     return model
 
@@ -798,13 +892,26 @@ def _optimizer_fn_to_optimizer(optimizer_fn: Union[Callable, None], model: Model
     optimizer = None
     if optimizer_fn:
         if framework == "tf":
-            optimizer = optimizer_fn()
+            try:
+                optimizer = optimizer_fn()
+            except:
+                raise AssertionError("optimizer_fn of Tensorflow backend should be callable without args. Please "
+                                     "make sure model and optimizer_fn are using the same backend")
+
             # initialize optimizer variables
             _ = optimizer.iterations
             optimizer._create_hypers()
             optimizer._create_slots(model.trainable_variables)
-            assert isinstance(optimizer, tf.optimizers.Optimizer)
+            # handle mixed precision loss scaling
+            if mixed_precision_tf.global_policy().name != "float32":
+                optimizer = mixed_precision_tf.LossScaleOptimizer(optimizer, loss_scale='dynamic')
+            assert isinstance(optimizer, tf.optimizers.Optimizer), "optimizer_fn should generate tensorflow optimizer"
         else:
-            optimizer = optimizer_fn(model.parameters())
-            assert isinstance(optimizer, torch.optim.Optimizer)
+            try:
+                optimizer = optimizer_fn(model.parameters())
+            except Exception as e:
+                print("optimizer_fn of Pytorch backend should be callable with single arg. Please sure model and \
+                optimizer_fn are using the same backend")
+                raise ValueError(repr(e))
+            assert isinstance(optimizer, torch.optim.Optimizer), "optimizer_fn should generate pytorch optimizer"
     return optimizer
