@@ -12,9 +12,10 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 # ==============================================================================
-from fastestimator.op.tensorop import loss, model, gradient
+from fastestimator.op.tensorop import augmentation, gradient, loss, model
 from fastestimator.op.tensorop.argmax import Argmax
 from fastestimator.op.tensorop.average import Average
 from fastestimator.op.tensorop.gather import Gather
 from fastestimator.op.tensorop.reshape import Reshape
-from fastestimator.op.tensorop.tensorop import TensorOp
+from fastestimator.op.tensorop.tensorop import LambdaOp, TensorOp
+from fastestimator.op.tensorop.un_hadamard import UnHadamard
