@@ -14,25 +14,36 @@
 # ==============================================================================
 import datetime
 import json
-from typing import Any, Optional
+from typing import Any, List, Optional, TYPE_CHECKING, Union
 
 import torch
 
 from fastestimator.network import BaseNetwork
+from fastestimator.pipeline import Pipeline
+from fastestimator.schedule.schedule import Scheduler
 from fastestimator.summary.summary import Summary
+from fastestimator.util.traceability_util import FeSummaryTable
+
+if TYPE_CHECKING:
+    from fastestimator.trace.trace import Trace
 
 
 class System:
     """A class which tracks state information while the fe.Estimator is running.
 
+    This class is intentionally not @traceable.
+
     Args:
         network: The network instance being used by the current fe.Estimator.
+        pipeline: The pipeline instance being used by the current fe.Estimator.
+        traces: The traces provided to the current fe.Estimator.
         mode: The current execution mode (or None for warmup).
         num_devices: How many GPUs are available for training.
         log_steps: Log every n steps (0 to disable train logging, None to disable all logging).
         total_epochs: How many epochs training is expected to run for.
-        max_train_steps_per_epoch: Whether training epochs will be cut short after N steps (or use None if they will run to
-            completion)
+        max_train_steps_per_epoch: Whether training epochs will be cut short after N steps (or use None if they will run
+            to completion)
+        system_config: A description of the initialization parameters defining the associated estimator.
 
     Attributes:
         mode: What is the current execution mode of the estimator ('train', 'eval', 'test'), None if warmup.
@@ -43,7 +54,9 @@ class System:
         epoch_idx: The current epoch index for the training (starting from 1).
         batch_idx: The current batch index within an epoch (starting from 1).
         stop_training: A flag to signal that training should abort.
-        network: A reference to the network being used this epoch
+        network: A reference to the network being used.
+        pipeline: A reference to the pipeline being used.
+        traces: The traces being used.
         max_train_steps_per_epoch: Training will complete after n steps even if loader is not yet exhausted.
         max_eval_steps_per_epoch: Evaluation will complete after n steps even if loader is not yet exhausted.
         summary: An object to write experiment results to.
@@ -59,6 +72,8 @@ class System:
     batch_idx: Optional[int]
     stop_training: bool
     network: BaseNetwork
+    pipeline: Pipeline
+    traces: List[Union['Trace', Scheduler['Trace']]]
     max_train_steps_per_epoch: Optional[int]
     max_eval_steps_per_epoch: Optional[int]
     summary: Summary
@@ -66,14 +81,19 @@ class System:
 
     def __init__(self,
                  network: BaseNetwork,
+                 pipeline: Pipeline,
+                 traces: List[Union['Trace', Scheduler['Trace']]],
                  mode: Optional[str] = None,
                  num_devices: int = torch.cuda.device_count(),
                  log_steps: Optional[int] = None,
                  total_epochs: int = 0,
                  max_train_steps_per_epoch: Optional[int] = None,
-                 max_eval_steps_per_epoch: Optional[int] = None) -> None:
+                 max_eval_steps_per_epoch: Optional[int] = None,
+                 system_config: Optional[List[FeSummaryTable]] = None) -> None:
 
         self.network = network
+        self.pipeline = pipeline
+        self.traces = traces
         self.mode = mode
         self.num_devices = num_devices
         self.log_steps = log_steps
@@ -82,7 +102,7 @@ class System:
         self.max_train_steps_per_epoch = max_train_steps_per_epoch
         self.max_eval_steps_per_epoch = max_eval_steps_per_epoch
         self.stop_training = False
-        self.summary = Summary(None)
+        self.summary = Summary(None, system_config)
         self.experiment_time = ""
         self._initialize_state()
 
@@ -130,18 +150,19 @@ class System:
         else:
             self.batch_idx += 1
 
-    def reset(self, summary_name: Optional[str] = None) -> None:
+    def reset(self, summary_name: Optional[str] = None, system_config: Optional[str] = None) -> None:
         """Reset the current `System` for a new round of training, including a new `Summary` object.
 
         Args:
             summary_name: The name of the experiment. The `Summary` object will store information iff name is not None.
+            system_config: A description of the initialization parameters defining the associated estimator.
         """
         self.experiment_time = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
         self.mode = "train"
         self._initialize_state()
         self.batch_idx = None
         self.stop_training = False
-        self.summary = Summary(summary_name)
+        self.summary = Summary(summary_name, system_config)
 
     def reset_for_test(self, summary_name: Optional[str] = None) -> None:
         """Partially reset the current `System` object for a new round of testing.
