@@ -30,8 +30,6 @@ from matplotlib.collections import PatchCollection
 from matplotlib.patches import Rectangle
 from pyfiglet import Figlet
 
-from fastestimator.backend.to_number import to_number
-
 STRING_TO_TORCH_DTYPE = {
     None: None,
     'float32': torch.float32,
@@ -51,6 +49,55 @@ STRING_TO_TORCH_DTYPE = {
     'bool': torch.bool
 }
 
+STRING_TO_TF_DTYPE = {
+    None: None,
+    "string": tf.string,
+    "int8": tf.int8,
+    "uint8": tf.uint8,
+    "int16": tf.int16,
+    "uint16": tf.uint16,
+    "int32": tf.int32,
+    "uint32": tf.uint32,
+    "int64": tf.int64,
+    "uint64": tf.uint64,
+    "float16": tf.float16,
+    "float32": tf.float32,
+    "float64": tf.float64
+}
+
+TENSOR_TO_NP_DTYPE = {
+    # Abstract types like 'float' and 'long' are intentionally not included here since they are never actually a
+    # tensor's dtype and they interfere with the finer-grained keys (torch.float intercepts torch.float32, for example)
+    None: None,
+    torch.float32: np.float32,
+    torch.float64: np.float64,
+    torch.float16: np.float16,
+    torch.uint8: np.uint8,
+    torch.int8: np.int8,
+    torch.int16: np.int16,
+    torch.int32: np.int32,
+    torch.int64: np.int64,
+    torch.bool: np.bool,
+    tf.float32: np.float32,
+    tf.float64: np.float64,
+    tf.float16: np.float16,
+    tf.uint8: np.uint8,
+    tf.int8: np.int8,
+    tf.int16: np.int16,
+    tf.int32: np.int32,
+    tf.int64: np.int64,
+    tf.bool: np.bool,
+    np.dtype('float32'): np.float32,
+    np.dtype('float64'): np.float64,
+    np.dtype('float16'): np.float16,
+    np.dtype('uint8'): np.uint8,
+    np.dtype('int8'): np.int8,
+    np.dtype('int16'): np.int16,
+    np.dtype('int32'): np.int32,
+    np.dtype('int64'): np.int64,
+    np.dtype('bool'): np.bool,
+}
+
 Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor)
 
 
@@ -113,7 +160,7 @@ def to_set(data: Any) -> Set[Any]:
     """Convert data to a set. A single None value will be converted to the empty set.
 
     ```python
-    x = fe.util.to_set(None)  # {}
+    x = fe.util.to_set(None)  # set()
     x = fe.util.to_set([None])  # {None}
     x = fe.util.to_set(7)  # {7}
     x = fe.util.to_set([7, 8])  # {7,8}
@@ -140,6 +187,8 @@ def to_set(data: Any) -> Set[Any]:
 class NonContext(object):
     """A class which is used to make nothing unusual happen.
 
+    This class is intentionally not @traceable.
+
     ```python
     a = 5
     with fe.util.NonContext():
@@ -157,6 +206,8 @@ class NonContext(object):
 class Suppressor(object):
     """A class which can be used to silence output of function calls.
 
+    This class is intentionally not @traceable.
+
     ```python
     x = lambda: print("hello")
     x()  # "hello"
@@ -186,9 +237,40 @@ class Suppressor(object):
         pass
 
 
+class LogSplicer:
+    """A class to send stdout information into a file before passing it along to the normal stdout.
+
+    Args:
+        log_path: The path/filename into which to append the current stdout.
+    """
+    def __init__(self, log_path: str):
+        self.log_path = log_path
+        self.stdout = None
+        self.log_file = None
+
+    def __enter__(self) -> None:
+        self.log_file = open(self.log_path, 'a')
+        self.stdout = sys.stdout
+        sys.stdout = self
+
+    def __exit__(self, *exc: Tuple[Optional[Type], Optional[Exception], Optional[Any]]) -> None:
+        sys.stdout = self.stdout
+        self.log_file.close()
+
+    def write(self, output: str) -> None:
+        self.log_file.write(output)
+        self.stdout.write(output)
+
+    def flush(self) -> None:
+        self.stdout.flush()
+        self.log_file.flush()
+
+
 class Timer(ContextDecorator):
     """A class that can be used to time things.
 
+    This class is intentionally not @traceable.
+
     ```python
     x = lambda: list(map(lambda i: i + i/2, list(range(int(1e6)))))
     with fe.util.Timer():
@@ -294,6 +376,15 @@ def get_type(obj: Any) -> str:
     x = fe.util.get_type(27)  # "int"
     ```
 
+    For container to look into its element's type, its type needs to be either list or tuple, and the return string will
+    be List[...]. All container elements need to have the same data type becuase it will only check its first element.
+
+    ```python
+    x = fe.util.get_type({"a":1, "b":2})  # "dict"
+    x = fe.util.get_type([1, "a"]) # "List[int]"
+    x = fe.util.get_type([[[1]]]) # "List[List[List[int]]]"
+    ```
+
     Args:
         obj: Data which may be wrapped in some kind of container.
 
@@ -383,13 +474,14 @@ def parse_modes(modes: Set[str]) -> Set[str]:
     return modes
 
 
-def pad_batch(batch: List[MutableMapping[str, Any]], pad_value: Union[float, int]) -> None:
-    """A function to pad a batch of data in-place by appending to the ends of the tensors.
+def pad_batch(batch: List[MutableMapping[str, np.ndarray]], pad_value: Union[float, int]) -> None:
+    """A function to pad a batch of data in-place by appending to the ends of the tensors. Tensor type needs to be
+    numpy array otherwise would get ignored. (tf.Tensor and torch.Tensor will cause error)
 
     ```python
     data = [{"x": np.ones((2, 2)), "y": 8}, {"x": np.ones((3, 1)), "y": 4}]
     fe.util.pad_batch(data, pad_value=0)
-    print(data)  # [{'x': [[1., 1.], [1., 1.],[0., 0.]], 'y': 8}, {'x': [[1., 0.], [1., 0.], [1., 0.]]), 'y': 4}]
+    print(data)  # [{'x': [[1., 1.], [1., 1.], [0., 0.]], 'y': 8}, {'x': [[1., 0.], [1., 0.], [1., 0.]]), 'y': 4}]
     ```
 
     Args:
@@ -397,9 +489,13 @@ def pad_batch(batch: List[MutableMapping[str, Any]], pad_value: Union[float, int
         pad_value: The value to pad with.
 
     Raises:
-        AssertionError: If the data within the batch do not have matching ranks.
+        AssertionError: If the data within the batch do not have matching rank, or have different keys
     """
-    for key in batch[0].keys():
+    keys = batch[0].keys()
+    for one_batch in batch:
+        assert one_batch.keys() == keys, "data within batch must have same keys"
+
+    for key in keys:
         shapes = [data[key].shape for data in batch if hasattr(data[key], "shape")]
         if len(set(shapes)) > 1:
             assert len(set(len(shape) for shape in shapes)) == 1, "data within batch must have same rank"
@@ -409,11 +505,14 @@ def pad_batch(batch: List[MutableMapping[str, Any]], pad_value: Union[float, int
 
 
 def pad_data(data: np.ndarray, target_shape: Tuple[int, ...], pad_value: Union[float, int]) -> np.ndarray:
-    """Pad `data` by appending `pad_value`s along it's dimensions until the `target_shape` is reached.
+    """Pad `data` by appending `pad_value`s along it's dimensions until the `target_shape` is reached. All entris of
+    target_shape should be larger than the data.shape, and have the same rank.
 
     ```python
     x = np.ones((1,2))
     x = fe.util.pad_data(x, target_shape=(3, 3), pad_value = -2)  # [[1, 1, -2], [-2, -2, -2], [-2, -2, -2]]
+    x = fe.util.pad_data(x, target_shape=(3, 3, 3), pad_value = -2) # error
+    x = fe.util.pad_data(x, target_shape=(4, 1), pad_value = -2) # error
     ```
 
     Args:
@@ -454,6 +553,8 @@ def is_number(arg: str) -> bool:
 class DefaultKeyDict(dict):
     """Like collections.defaultdict but it passes the key argument to the default function.
 
+    This class is intentionally not @traceable.
+
     ```python
     d = fe.util.DefaultKeyDict(default=lambda x: x+x, a=4, b=6)
     print(d["a"])  # 4
@@ -488,16 +589,42 @@ def show_image(im: Union[np.ndarray, Tensor],
                title: Optional[str] = None,
                color_map: str = "inferno",
                stack_depth: int = 0) -> Optional[plt.Figure]:
-    """Plots a given image onto an axis.
+    """Plots a given image onto an axis. The repeated invocation of this function will cause figure plot overlap.
+
+    If `im` is 2D and the length of second dimension are 4 or 5, it will be viewed as bounding box data (x0, y0, w, h,
+    <label>).
+
+    ```python
+    boxes = np.array([[0, 0, 10, 20, "apple"],
+                      [10, 20, 30, 50, "dog"],
+                      [40, 70, 200, 200, "cat"],
+                      [0, 0, 0, 0, "not_shown"],
+                      [0, 0, -10, -20, "not_shown2"]])
+
+    img = np.zeros((150, 150))
+    fig, axis = plt.subplots(1, 1)
+    fe.util.show_image(img, fig=fig, axis=axis) # need to plot image first
+    fe.util.show_image(boxes, fig=fig, axis=axis)
+    ```
+
+    Users can also directly plot text
+
+    ```python
+    fig, axis = plt.subplots(1, 1)
+    fe.util.show_image("apple", fig=fig, axis=axis)
+    ```
 
     Args:
         axis: The matplotlib axis to plot on, or None for a new plot.
         fig: A reference to the figure to plot on, or None if new plot.
-        im: The image to display (width X height).
+        im: The image (width X height) / bounding box / text to display.
         title: A title for the image.
         color_map: Which colormap to use for greyscale images.
         stack_depth: Multiple images can be drawn onto the same axis. When stack depth is greater than zero, the `im`
             will be alpha blended on top of a given axis.
+
+    Returns:
+        plotted figure. It will be the same object as user have provided in the argument.
     """
     if axis is None:
         fig, axis = plt.subplots(1, 1)
@@ -537,8 +664,8 @@ def show_image(im: Union[np.ndarray, Tensor],
             height = int(box[3])
             label = None if len(box) < 5 else str(box[4])
 
-            # Don't draw empty boxes
-            if width == 0 and height == 0:
+            # Don't draw empty boxes, or invalid box
+            if width <= 0 or height <= 0:
                 continue
             r = Rectangle((x0, y0), width=width, height=height, fill=False, edgecolor=color, linewidth=3)
             boxes.append(r)
@@ -549,7 +676,7 @@ def show_image(im: Union[np.ndarray, Tensor],
                           ha='left',
                           va='top',
                           color=color,
-                          fontsize=min(14, width // len(label)),
+                          fontsize=max(8, min(14, width // len(label))),
                           fontweight='bold',
                           family='monospace')
         pc = PatchCollection(boxes, match_original=True)
@@ -562,12 +689,16 @@ def show_image(im: Union[np.ndarray, Tensor],
             im = im.permute(*channels)
         # image data
         im = to_number(im)
+        im_max = np.max(im)
+        im_min = np.min(im)
         if np.issubdtype(im.dtype, np.integer):
             # im is already in int format
             im = im.astype(np.uint8)
-        elif np.max(im) <= 1 and np.min(im) >= 0:  # im is [0,1]
+        elif 0 <= im_min <= im_max <= 1:  # im is [0,1]
             im = (im * 255).astype(np.uint8)
-        elif np.min(im) >= -1 and np.max(im) <= 1:  # im is [-1, 1]
+        elif -0.5 <= im_min < 0 < im_max <= 0.5:  # im is [-0.5, 0.5]
+            im = ((im + 0.5) * 255).astype(np.uint8)
+        elif -1 <= im_min < 0 < im_max <= 1:  # im is [-1, 1]
             im = ((im + 1) * 127.5).astype(np.uint8)
         else:  # im is in some arbitrary range, probably due to the Normalize Op
             ma = abs(np.max(im, axis=tuple([i for i in range(len(im.shape) - 1)]) if len(im.shape) > 2 else None))
@@ -587,7 +718,8 @@ def show_image(im: Union[np.ndarray, Tensor],
 
 
 def get_batch_size(data: Dict[str, Any]) -> int:
-    """Infer batch size from a batch dictionary.
+    """Infer batch size from a batch dictionary. It will ignore all dictionary value with data type that
+    doesn't have "shape" attribute.
 
     Args:
         data: The batch dictionary.
@@ -599,3 +731,110 @@ def get_batch_size(data: Dict[str, Any]) -> int:
     batch_size = set(data[key].shape[0] for key in data if hasattr(data[key], "shape") and list(data[key].shape))
     assert len(batch_size) == 1, "invalid batch size: {}".format(batch_size)
     return batch_size.pop()
+
+
+class FEID:
+    """An int wrapper class that can change how it's values are printed.
+
+    This class is intentionally not @traceable.
+
+    Args:
+        val: An integer id to be wrapped.
+    """
+    __slots__ = ['_val']
+    _translation_dict = {}
+
+    def __init__(self, val: int):
+        self._val = val
+
+    def __hash__(self) -> int:
+        return hash(self._val)
+
+    def __eq__(self, other: Any) -> bool:
+        if isinstance(other, FEID):
+            return self._val == other._val
+        else:
+            return int.__eq__(self._val, other)
+
+    def __lt__(self, other: Any) -> bool:
+        if isinstance(other, FEID):
+            other = other._val
+        return int.__lt__(self._val, other)
+
+    def __str__(self) -> str:
+        return f"{self._translation_dict.get(self._val, self._val)}"
+
+    def __repr__(self) -> str:
+        return f"{self._translation_dict.get(self._val, self._val)}"
+
+    @classmethod
+    def set_translation_dict(cls, mapping: Dict[int, Any]) -> None:
+        """Provide a lookup table to be invoked during value printing.
+
+        Args:
+            mapping: A mapping of id: printable id.
+        """
+        cls._translation_dict.clear()
+        cls._translation_dict.update(mapping)
+
+
+class Flag:
+    """A mutable wrapper around a boolean.
+
+    This class is intentionally not @traceable.
+
+    Args:
+        val: The initial value for the Flag.
+    """
+    __slots__ = ['_val']
+
+    def __init__(self, val: bool = False):
+        self._val = val
+
+    def set_true(self):
+        self._val = True
+
+    def set_false(self):
+        self._val = False
+
+    def __bool__(self):
+        return self._val
+
+
+def to_number(data: Union[tf.Tensor, torch.Tensor, np.ndarray, int, float]) -> np.ndarray:
+    """Convert an input value into a Numpy ndarray.
+
+    This method can be used with Python and Numpy data:
+    ```python
+    b = fe.backend.to_number(5)  # 5 (type==np.ndarray)
+    b = fe.backend.to_number(4.0)  # 4.0 (type==np.ndarray)
+    n = np.array([1, 2, 3])
+    b = fe.backend.to_number(n)  # [1, 2, 3] (type==np.ndarray)
+    ```
+
+    This method can be used with TensorFlow tensors:
+    ```python
+    t = tf.constant([1, 2, 3])
+    b = fe.backend.to_number(t)  # [1, 2, 3] (type==np.ndarray)
+    ```
+
+    This method can be used with PyTorch tensors:
+    ```python
+    p = torch.tensor([1, 2, 3])
+    b = fe.backend.to_number(p)  # [1, 2, 3] (type==np.ndarray)
+    ```
+
+    Args:
+        data: The value to be converted into a np.ndarray.
+
+    Returns:
+        An ndarray corresponding to the given `data`.
+    """
+    if tf.is_tensor(data):
+        data = data.numpy()
+    elif isinstance(data, torch.Tensor):
+        if data.requires_grad:
+            data = data.detach().numpy()
+        else:
+            data = data.numpy()
+    return np.array(data)
